# This file is placed in the Public Domain.


"clients"


import queue
import threading
import _thread


from .brokers import Fleet
from .handler import Handler
from .threads import launch


class Client(Handler):

    def __init__(self):
        Handler.__init__(self)
        self.olock = threading.RLock()
        self.oqueue = queue.Queue()
        self.silent = True
        Fleet.add(self)

    def announce(self, txt):
        if not self.silent:
            self.raw(txt)

    def display(self, event):
        with self.olock:
            for tme in sorted(event.result):
                self.dosay(
                           event.channel,
                           event.result[tme]
                          )

    def dosay(self, channel, txt):
        self.say(channel, txt)

    def raw(self, txt):
        raise NotImplementedError("raw")

    def say(self, channel, txt):
        self.raw(txt)


class Output(Client):

    def output(self):
        while True:
            event = self.oqueue.get()
            if event is None:
                self.oqueue.task_done()
                break
            self.display(event)
            self.oqueue.task_done()

    def start(self):
        launch(self.output)
        super().start()

    def stop(self):
        self.oqueue.put(None)
        super().stop()

    def wait(self):
        try:
            self.oqueue.join()
        except Exception:
            _thread.interrupt_main()


def __dir__():
    return (
        'Client',
        'Output'
   )



# This file is placed in the Public Domain.


"wijs, wijsheid, wijs!"


import logging


from random import SystemRandom


from tob.brokers import Fleet
from tob.handler import Event
from tob.threads import Repeater


rand = SystemRandom()


def init():
    event = Event()
    repeater = Repeater(3600.0,  wsd, event)
    repeater.start()
    logging.warning("%s wise", len(TXT.split("\n")))


def wsd(event):
    Fleet.announce(rand.choice(TXT.split("\n")).strip()[2:])


TXT = """| wijsheid, wijs !

| OVERDRACHT
| ==========

| No Voice, No Description:

| enkelt eenlingen
| knipperogen
| ARC
| boot dromerij (mr_boeking)
| kwellerei
| schuldwijzerij
| klappenknijperij
| heelalknijperij
| de afleggingen
| dag
| oordeel
| zak door de grond (wederopwekking test, test = roep je er om)
| passage, voorzetterij, vang maar op
| projectie
| nieuw making (het vervolg leggende)
| hemel zakking
| ok computer
| nieuwe hemel
| nieuw buiten heersende
| dag sluiting
| uurtjes
| nazi met hel zoekerij
| peace departement
| muur stempelen
| muur duwerij
| eeuwige oordeel analyse (het toon jezelf)
| nieuwe zachtaardige
| goed oefening (inspirate geving)
| herkenning van oud goed
| woorden, woorden
| WOORD
| droom analyse
| ballen, werelden
| zone drukking
| enkelingen over
| in boek
| steek en eerste lichting evaluatie
| domeintjes
| les brenging (hemel poorten, hell legging)
| dubbelzinning praat
| mensen in den hoge
| ter verantwoording roeping (het goede woord)
| hulp roeping (melding)
| woord herkenning door licht (het begrip)
| tripping (de auto)
| mental sex (enorm lang lopen geweest, met 1e schuld legging erbij)
| tweede schuldlegging is goed gedaan op het atomaire vlak (gebied)
| derde schuldlegging is materiaal tot enkelingen gereduceerd
| vierde schuldlegging is de terug schuldlegging (schepping mishandelt)
| onschuld
| oordeel is hier het gesproken woord
| woord gebruik
| stil makerij
| constanten
| vroeger onzichtbare
| De Mens
| De Hemel
| pad terug gegeven (en nooit het gegeven pad nemen)
| iets nemen is dan een kiezen (tandarts)
| crews (besturing)
| levenden in materiaal
| levenden in hell (uitgebeeld)
| OVERDRACHT
| company, mens beheer op aarde, ziet atlen komen en gaan
| poging tot nieuw start
| de arc types (hier selecteren om einde aan de rest te noemen)
| boven gelegd woord (zeewier in vissenbak)
| het 4 leg zysteem (woord gelegd op schepping .. voor making)
| de immer zijnde dreiging tot einde
| de weigerende mens hier toe (de weerwil)
| scheppers vragen
| instappend beeld
| judgement has passed
| ring (wit/rood) als mr. evidence calling
| na de komma als uitleg, buiten lopend reciterende
| atlje coden
| eigendom discussies
| kooi houderij
| in mental verhaal houderij
| laat die 2 maar alleen (de boodschappers)
| the film is the prison
| hell als gevangenis
| oordeel millenia sluiting
| vrijlaten van alle gevangenen akkoord
| dit alles tot einde oordeel millenium
| nog meer om meneer echt niet op de knop te laten drukken
| de vragende mens, die vraagt om nog meer vragen
| trotse uitbeelding, verwoording
| constante correctheids checking
| overdracht met zonde melding
| alice trap naar de hemel uitgebeeld
| buiten koppeling getoond
| de mens zijn gemanipuleerde fantasie gebruikt als trigger van buiten fenomeen
| ook weer als fenomenen getoond
| controle over fenomenen getoond
| levenden in droom vragen om aandacht en melding
| dagjes wel of niet bij mensen in de droom
| boven hemelen vermelding
| kapot schopperij en hoe lang dat geduurt heeft
| melding voor terug gedaan om melding te ontdoen van zijn kracht
| vergeten, vergeten
| suffering on the floor
| schuld ontkenning als de tergkracht
| stekker verhaal, afscheid gabriel
| droom geheugen en reeel geheugen terug voeding
| woordleggers
| stroom tagging
| attributen beelding (twee op de schouders)
| boeken referentie, opgelegd woord als proeve
| moraal roeping
| haakse als effectief op gedachtenstroom (uitleg van de hemel)
| de een zijn hemel is de ander zijn hel
| 'het woord begeleid', nieuw komenden
| ultra addaptables
| terminator badjes
| denk instraal met materiaal (zonder handjes)
| huidig bepaling
| engine (gruis)
| essentiele vergeetachtigheid
| mental working (mogelijkheden schepping)
| attributen bepaling
| droom insluiting
| diamanten aan de onderkant van de boot
| zware oogst in de hemelen
| dag van vandaag coden (random op gegeven topics)
| oude pijn ignored
| nog niet gespoelt (nog niet verkeerd afgerond)
| de dingen liggen nog open
| stilstaande stroom (water)
| plasma
| groen-zwart acid burn
| 4 kwadranten bijna opening
| testen op de geest
| oog voorbijgaand (passerend)
| genezing belofte (het herstelt zich vanzelf)
| muziek bezoektes
| schoonmakerij
| de persoonoogst
| caretaker discussions (double their wages)
| de verkeerd noeming
| algemeenheids denken getoond (laagje erachter)
| legacy vernoemd
| de heer van de beheersing heerst hier, kuisheid
| schuld projectie is hoeks erop, uitslaan en uit besef doorslaan
| dreig projectie, torque .. de armdraai
| haaks erop de heen en weer dans, actie/reactie met verkeerd en goed vermeerderen
| false profeet omklap (linker pop die van meekijken naar inkijken woord)
| de dag na dag deze dag (vandaag)
| meervoudige oordeels voering
| materiaal reactie eromheen
| putten dempers
| zelfvertoning, duidbaar gemaakt worden
| de belofte man is er weer
| slecht advies van buiten melding
| constante redders nood melding
| de overname mensen zeggen er geen last van te hebben
| topic naming als presence
| engels/nederlands als vroegtijde dag voorkomers
| strictly dutch as The Day Preventers
| corner bouncing
| long distance woord over de ruit
| stiltewens genoemd
| de doorpraten noodzaak
| stok sturende met niet gemaaktheid als gezien (opgemerkt)
| tagging/merking is de moeilijke zaak
| schuldwijzing is de gaande zaak
| terugwijzer als immer aanwezige
| niet op mensen, niet op vogels
| brievenbus en graf
| twee eieren (gemeente huis en buiging naar west)
| vreedevernoeming
| duiding
| coding
"""



# This file is placed in the Public Domain.


"tob"


import unittest


from tob.objects import Object, items, keys, update, values


import tob.objects


OBJECT = Object()
PACKAGE = tob.objects
VALIDJSON = "{'test': 'bla'}"
VALIDPYTHON = '{"test": "bla"}'


class Mix:
    a = "b"


class Mixin(Mix, Object):
    pass


attrs1 = (
    "Object",
    "construct",
    "dumps",
    "items",
    "keys",
    "loads",
    "update",
    "values"
)


attrs2 = (
    "__doc__",
    "__lt__",
    "__init__",
    "__setattr__",
    "__ne__",
    "__delattr__",
    "__eq__",
    "__dir__",
    "__new__",
    "__iter__",
    "__reduce__",
    "__class__",
    "__module__",
    "__gt__",
    "__str__",
    "__init_subclass__",
    "__reduce_ex__",
    "__dict__",
    "__subclasshook__",
    "__le__",
    "__contains__",
    "__weakref__",
    "__ge__",
    "__sizeof__",
    "__getattribute__",
    "__format__",
    "__len__",
    "__getstate__",
    "__repr__",
    "__hash__",
)


class TestObject(unittest.TestCase):

    def test_attributes(self):
        okd = True
        for meth in attrs2:
            mth = getattr(OBJECT, meth, None)
            if mth is None:
                okd = meth
        self.assertTrue(okd)

    def test_constructor(self):
        obj = Object()
        self.assertTrue(type(obj), Object)

    def test_class(self):
        obj = Object()
        clz = obj.__class__()
        self.assertTrue("Object" in str(type(clz)))

    def test_delattr(self):
        obj = Object()
        obj.key = "value"
        del obj.key
        self.assertTrue("key" not in dir(obj))

    def test_dict(self):
        obj = Object()
        self.assertEqual(obj.__dict__, {})

    def test_fmt(self):
        obj = Object()
        self.assertEqual(format(obj), "{}")

    def test_format(self):
        obj = Object()
        self.assertEqual(format(obj), "{}")

    def test_getattribute(self):
        obj = Object()
        obj.key = "value"
        self.assertEqual(obj.__getattribute__("key"), "value")

    def test_getattr(self):
        obj = Object()
        obj.key = "value"
        self.assertEqual(getattr(obj, "key"), "value")

    def test_hash(self):
        obj = Object()
        hsj = hash(obj)
        self.assertTrue(isinstance(hsj, int))

    def test_init(self):
        obj = Object()
        self.assertTrue(type(Object.__init__(obj)), Object)

    def test_items(self):
        obj = Object()
        obj.key = "value"
        self.assertEqual(
            list(items(obj)),
            [
                ("key", "value"),
            ],
        )

    def test_keys(self):
        obj = Object()
        obj.key = "value"
        self.assertEqual(
            list(keys(obj)),
            [
                "key",
            ],
        )

    def test_len(self):
        obj = Object()
        self.assertEqual(len(obj), 0)

    def test_methods(self):
        okd = True
        for attr in attrs1:
            att = getattr(PACKAGE, attr, None)
            if not att:
                okd = attr
                break
        self.assertTrue(okd)

    def test_module(self):
        self.assertEqual(Object().__module__, "tob.objects")

    def test_register(self):
        obj = Object()
        setattr(obj, "key", "value")
        self.assertEqual(obj.key, "value")

    def test_repr(self):
        self.assertTrue(update(Object(), {"key": "value"}).__repr__(), {"key": "value"})

    def test_setattr(self):
        obj = Object()
        obj.__setattr__("key", "value")
        self.assertTrue(obj.key, "value")

    def test_str(self):
        obj = Object()
        self.assertEqual(str(obj), "{}")

    def test_update(self):
        obj = Object()
        obj.key = "value"
        oobj = Object()
        update(oobj, obj)
        self.assertTrue(oobj.key, "value")

    def test_values(self):
        obj = Object()
        obj.key = "value"
        self.assertEqual(
            list(values(obj)),
            [
                "value",
            ],
        )



# This file is placed in the Public Domain.


"rich site syndicate"


import html
import html.parser
import http.client
import logging
import os
import re
import time
import urllib
import urllib.parse
import urllib.request
import uuid
import _thread


from urllib.error import HTTPError, URLError
from urllib.parse import quote_plus, urlencode


from tob.brokers import Fleet
from tob.objects import Object, update
from tob.persist import find, fntime, getpath, last, write
from tob.threads import Repeater, launch
from tob.utility import elapsed, fmt, spl


def init():
    fetcher = Fetcher()
    fetcher.start()
    if fetcher.seenfn:
        logging.warning("since %s", elapsed(time.time()-fntime(fetcher.seenfn)))
    else:
        logging.warning("since %s", time.ctime(time.time()))
    return fetcher


DEBUG = False


fetchlock = _thread.allocate_lock()
importlock = _thread.allocate_lock()
errors = {}
skipped = []


class Feed:

    def __init__(self):
        self.link = ""
        self.name = ""


class Rss:

    def __init__(self):
        self.display_list = "title,link,author"
        self.insertid = None
        self.name = ""
        self.rss = ""


class Urls:

    pass


class Fetcher(Object):

    def __init__(self):
        self.dosave = False
        self.seen = Urls()
        self.seenfn = None

    @staticmethod
    def display(obj):
        displaylist = ""
        result = ""
        try:
            displaylist = obj.display_list or "title,link"
        except AttributeError:
            displaylist = "title,link,author"
        for key in displaylist.split(","):
            if not key:
                continue
            data = getattr(obj, key, None)
            if not data:
                continue
            data = data.replace("\n", " ")
            data = striphtml(data.rstrip())
            data = unescape(data)
            result += data.rstrip()
            result += " - "
        return result[:-2].rstrip()

    def fetch(self, feed, silent=False):
        with fetchlock:
            result = []
            seen = getattr(self.seen, feed.rss, [])
            urls = []
            counter = 0
            for obj in reversed(getfeed(feed.rss, feed.display_list)):
                counter += 1
                fed = Feed()
                update(fed, obj)
                update(fed, feed)
                url = urllib.parse.urlparse(fed.link)
                if url.path and not url.path == "/":
                    uurl = f"{url.scheme}://{url.netloc}/{url.path}"
                else:
                    uurl = fed.link
                urls.append(uurl)
                if uurl in seen:
                    continue
                if self.dosave:
                    write(fed)
                result.append(fed)
            setattr(self.seen, feed.rss, urls)
            if not self.seenfn:
                self.seenfn = getpath(self.seen)
            write(self.seen, self.seenfn)
        if silent:
            return counter
        txt = ""
        feedname = getattr(feed, "name", None)
        if feedname:
            txt = f"[{feedname}] "
        for obj in result:
            txt2 = txt + self.display(obj)
            for bot in Fleet.all():
                bot.announce(txt2)
        return counter

    def run(self, silent=False):
        thrs = []
        for _fn, feed in find("rss.Rss"):
            thrs.append(launch(self.fetch, feed, silent))
        return thrs

    def start(self, repeat=True):
        self.seenfn = last(self.seen)
        if repeat:
            repeater = Repeater(300.0, self.run)
            repeater.start()


class Parser:

    @staticmethod
    def getitem(line, item):
        lne = ""
        index1 = line.find(f"<{item}>")
        if index1 == -1:
            return lne
        index1 += len(item) + 2
        index2 = line.find(f"</{item}>", index1)
        if index2 == -1:
            return lne
        lne = line[index1:index2]
        lne = cdata(lne)
        return lne.strip()

    @staticmethod
    def getitems(text, token):
        index = 0
        result = []
        stop = False
        while not stop:
            index1 = text.find(f"<{token}", index)
            if index1 == -1:
                break
            index1 += len(token) + 2
            index2 = text.find(f"</{token}>", index1)
            if index2 == -1:
                break
            lne = text[index1:index2]
            result.append(lne)
            index = index2
        return result

    @staticmethod
    def parse(txt, toke="item", items="title,link"):
        result = []
        for line in Parser.getitems(txt, toke):
            line = line.strip()
            obj = Object()
            for itm in spl(items):
                val = Parser.getitem(line, itm)
                if val:
                    val = unescape(val.strip())
                    val = val.replace("\n", "")
                    val = striphtml(val)
                    setattr(obj, itm, val)
            result.append(obj)
        return result


"OPML"


class OPML:

    @staticmethod
    def getnames(line):
        return [x.split('="')[0] for x in line.split()]

    @staticmethod
    def getvalue(line, attr):
        lne = ""
        index1 = line.find(f'{attr}="')
        if index1 == -1:
            return lne
        index1 += len(attr) + 2
        index2 = line.find('"', index1)
        if index2 == -1:
            index2 = line.find("/>", index1)
        if index2 == -1:
            return lne
        lne = line[index1:index2]
        if "CDATA" in lne:
            lne = lne.replace("![CDATA[", "")
            lne = lne.replace("]]", "")
            # lne = lne[1:-1]
        return lne

    @staticmethod
    def getattrs(line, token):
        index = 0
        result = []
        stop = False
        while not stop:
            index1 = line.find(f"<{token} ", index)
            if index1 == -1:
                return result
            index1 += len(token) + 2
            index2 = line.find("/>", index1)
            if index2 == -1:
                return result
            result.append(line[index1:index2])
            index = index2
        return result

    @staticmethod
    def parse(txt, toke="outline", itemz=None):
        if itemz is None:
            itemz = ",".join(OPML.getnames(txt))
        result = []
        for attrz in OPML.getattrs(txt, toke):
            if not attrz:
                continue
            obj = Object()
            for itm in spl(itemz):
                if itm == "link":
                    itm = "href"
                val = OPML.getvalue(attrz, itm)
                if not val:
                    continue
                if itm == "href":
                    itm = "link"
                setattr(obj, itm, val.strip())
            result.append(obj)
        return result


"utilities"


def attrs(obj, txt):
    update(obj, OPML.parse(txt))


def cdata(line):
    if "CDATA" in line:
        lne = line.replace("![CDATA[", "")
        lne = lne.replace("]]", "")
        lne = lne[1:-1]
        return lne
    return line


def getfeed(url, items):
    result = [Object(), Object()]
    if DEBUG or url in errors and (time.time() - errors[url]) < 600:
        return result
    try:
        rest = geturl(url)
    except (http.client.HTTPException, ValueError, HTTPError, URLError) as ex:
        logging.error("%s %s", url, ex)
        errors[url] = time.time()
        return result
    if rest:
        if "link" not in items:
            items += ",link"
        if url.endswith("atom"):
            result = Parser.parse(str(rest.data, "utf-8"), "entry", items) or []
        else:
            result = Parser.parse(str(rest.data, "utf-8"), "item", items) or []
    return result


def gettinyurl(url):
    postarray = [
        ("submit", "submit"),
        ("url", url),
    ]
    postdata = urlencode(postarray, quote_via=quote_plus)
    req = urllib.request.Request(
        "http://tinyurl.com/create.php", data=bytes(postdata, "UTF-8")
    )
    req.add_header("User-agent", useragent("rss fetcher"))
    with urllib.request.urlopen(req) as htm:  # nosec
        for txt in htm.readlines():
            line = txt.decode("UTF-8").strip()
            i = re.search('data-clipboard-text="(.*?)"', line, re.M)
            if i:
                return i.groups()
    return []


def geturl(url):
    url = urllib.parse.urlunparse(urllib.parse.urlparse(url))
    req = urllib.request.Request(str(url))
    req.add_header("User-agent", useragent("rss fetcher"))
    with urllib.request.urlopen(req) as response:  # nosec
        response.data = response.read()
        return response


def shortid():
    return str(uuid.uuid4())[:8]


def striphtml(text):
    clean = re.compile("<.*?>")
    return re.sub(clean, "", text)


def unescape(text):
    txt = re.sub(r"\s+", " ", text)
    return html.unescape(txt)


def useragent(txt):
    return "Mozilla/5.0 (X11; Linux x86_64) " + txt


"commands"


def dpl(event):
    if len(event.args) < 2:
        event.reply("dpl <stringinurl> <item1,item2>")
        return
    setter = {"display_list": event.args[1]}
    for fnm, feed in find("rss.Rss", {"rss": event.args[0]}):
        if feed:
            update(feed, setter)
            write(feed, fnm)
    event.reply("ok")


def exp(event):
    with importlock:
        event.reply(TEMPLATE)
        nrs = 0
        for _fn, ooo in find("rss.Rss"):
            nrs += 1
            obj = Rss()
            update(obj, ooo)
            name = f"url{nrs}"
            txt = f'<outline name="{name}" display_list="{obj.display_list}" xmlUrl="{obj.rss}"/>'
            event.reply(" " * 12 + txt)
        event.reply(" " * 8 + "</outline>")
        event.reply("    <body>")
        event.reply("</opml>")


def imp(event):
    if not event.args:
        event.reply("imp <filename>")
        return
    fnm = event.args[0]
    if not os.path.exists(fnm):
        event.reply(f"no {fnm} file found.")
        return
    with importlock:
        with open(fnm, "r", encoding="utf-8") as file:
            txt = file.read()
        prs = OPML()
        nrs = 0
        nrskip = 0
        insertid = shortid()
        for obj in prs.parse(txt, "outline", "name,display_list,xmlUrl"):
            url = obj.xmlUrl
            if url in skipped:
                continue
            if not url.startswith("http"):
                continue
            has = list(find("rss.Rss", {"rss": url}, matching=True))
            if has:
                skipped.append(url)
                nrskip += 1
                continue
            feed = Rss()
            update(feed, obj)
            feed.rss = obj.xmlUrl
            feed.insertid = insertid
            write(feed)
            nrs += 1
    if nrskip:
        event.reply(f"skipped {nrskip} urls.")
    if nrs:
        event.reply(f"added {nrs} urls.")


def nme(event):
    if len(event.args) != 2:
        event.reply("nme <stringinurl> <name>")
        return
    selector = {"rss": event.args[0]}
    for fnm, fed in find("rss.Rss", selector):
        feed = Rss()
        update(feed, fed)
        if feed:
            feed.name = str(event.args[1])
            write(feed, fnm)
    event.reply("ok")


def rem(event):
    if len(event.args) != 1:
        event.reply("rem <stringinurl>")
        return
    for fnm, fed in find("rss.Rss"):
        feed = Rss()
        update(feed, fed)
        if event.args[0] not in feed.rss:
            continue
        if feed:
            feed.__deleted__ = True
            write(feed, fnm)
            event.reply("ok")
            break


def res(event):
    if len(event.args) != 1:
        event.reply("res <stringinurl>")
        return
    for fnm, fed in find("rss.Rss", removed=True):
        feed = Rss()
        update(feed, fed)
        if event.args[0] not in feed.rss:
            continue
        if feed:
            feed.__deleted__ = False
            write(feed, fnm)
    event.reply("ok")


def rss(event):
    if not event.rest:
        nrs = 0
        for fnm, fed in find("rss.Rss"):
            nrs += 1
            elp = elapsed(time.time() - fntime(fnm))
            txt = fmt(fed)
            event.reply(f"{nrs} {txt} {elp}")
        if not nrs:
            event.reply("no feed found.")
        return
    url = event.args[0]
    if "http://" not in url and "https://" not in url:
        event.reply("i need an url")
        return
    for fnm, result in find("rss.Rss", {"rss": url}):
        if result:
            event.reply(f"{url} is known")
            return
    feed = Rss()
    feed.rss = event.args[0]
    write(feed)
    event.reply("ok")


def syn(event):
    if DEBUG:
        return
    fetcher = Fetcher()
    fetcher.start(False)
    thrs = fetcher.run(True)
    nrs = 0
    for thr in thrs:
        thr.join()
        nrs += 1
    event.reply(f"{nrs} feeds synced")


TEMPLATE = """<opml version="1.0">
    <head>
        <title>OPML</title>
    </head>
    <body>
        <outline title="opml" text="rss feeds">"""



# This file is placed in the Public Domain.


"show path to website"


import os


a = os.path.abspath
d = os.path.dirname
p = os.path.join


PATH = p(d(d(__file__)), "network", "index.html")


def pth(event):
    event.reply(f"file://{PATH}")



# This file is placed in the Public Domain.


"handle events"


import queue
import threading
import time
import _thread


from .threads import launch


class Event:

    def __init__(self):
        self._ready = threading.Event()
        self._thr = None
        self.channel = ""
        self.ctime = time.time()
        self.orig = ""
        self.result = {}
        self.txt = ""
        self.type = "event"

    def ready(self):
        self._ready.set()

    def reply(self, txt):
        self.result[time.time()] = txt

    def wait(self, timeout=None):
        try:
            self._ready.wait()
            if self._thr:
                self._thr.join(timeout)
        except (KeyboardInterrupt, EOFError):
            _thread.interrupt_main()


class Handler:

    def __init__(self):
        self.cbs = {}
        self.queue = queue.Queue()

    def callback(self, event):
        func = self.cbs.get(event.type, None)
        if func:
            name = event.txt and event.txt.split()[0]
            event._thr = launch(func, event, name=name)
        else:
            event.ready()

    def loop(self):
        while True:
            try:
                event = self.poll()
                if event is None:
                    break
                event.orig = repr(self)
                self.callback(event)
            except (KeyboardInterrupt, EOFError):
                _thread.interrupt_main()

    def poll(self):
        return self.queue.get()

    def put(self, event):
        self.queue.put(event)

    def register(self, type, callback):
        self.cbs[type] = callback

    def start(self):
        launch(self.loop)

    def stop(self):
        self.queue.put(None)


def __dir__():
    return (
        'Event',
        'Handler'
   )



# This file is placed in the Public Domain.


"write your own commands"


import inspect


from .brokers import Fleet
from .package import getmod, modules
from .utility import parse


class Commands:

    cmds = {}
    names = {}

    @staticmethod
    def add(*args) -> None:
        for func in args:
            name = func.__name__
            Commands.cmds[name] = func
            Commands.names[name] = func.__module__.split(".")[-1]

    @staticmethod
    def get(cmd):
        func = Commands.cmds.get(cmd, None)
        if func:
            return func
        name = Commands.names.get(cmd, None)
        if name:
            module = getmod(name)
            if module:
                scan(module)
        return Commands.cmds.get(cmd, None)


def command(evt):
    parse(evt, evt.txt)
    func = Commands.get(evt.cmd)
    if func:
        func(evt)
        Fleet.display(evt)
    evt.ready()


def scan(module):
    for key, cmdz in inspect.getmembers(module, inspect.isfunction):
        if key.startswith("cb"):
            continue
        if 'event' in inspect.signature(cmdz).parameters:
            Commands.add(cmdz)


def scanner(names=[]):
    res = []
    for nme in modules():
        if names and nme not in names:
            continue
        module = getmod(nme)
        if not module:
            continue
        scan(module)
        res.append(module)
    return res


def table():
    tbl = getmod("tbl")
    if tbl and "NAMES" in dir(tbl):
        Commands.names.update(tbl.NAMES)
    else:
        scanner()


def __dir__():
    return (
        'Comamnds',
        'command',
        'parse',
        'scan',
        'scanner',
        'table'
    )



# This file is placed in the Public Domain.


"module management"


import importlib
import importlib.util
import logging
import os
import sys
import _thread


from .threads import launch
from .utility import md5sum


class Mods:

    dirs = {}
    md5s = {}

    def add(name, path):
        Mods.dirs[name] = path


def getmod(name):
    for nme, path in Mods.dirs.items():
        mname = nme + "." +  name
        module = sys.modules.get(mname, None)
        if module:
            return module
        pth = os.path.join(path, f"{name}.py")
        if Mods.md5s:
            if os.path.exists(pth) and name != "tbl":
                md5 = Mods.md5s.get(name, None)
                if md5sum(pth) != md5:
                    logging.info("md5 error %s", name)
        mod = importer(mname, pth)
        if mod:
            return mod


def importer(name, pth):
    if not os.path.exists(pth):
        return
    try:
        spec = importlib.util.spec_from_file_location(name, pth)
        if not spec or not spec.loader:
            return
        mod = importlib.util.module_from_spec(spec)
        if not mod:
            return
        sys.modules[name] = mod
        spec.loader.exec_module(mod)
        return mod
    except Exception as ex:
        logging.exception(ex)
        _thread.interrupt_main()


def inits(names):
    modz = []
    for name in modules():
        if name not in names:
            continue
        try:
            module = getmod(name)
            if module and "init" in dir(module):
                thr = launch(module.init)
                modz.append((module, thr))
        except Exception as ex:
            logging.exception(ex)
            _thread.interrupt_main()
    return modz


def modules():
    mods = []
    for name, path in Mods.dirs.items():
        if not os.path.exists(path):
            continue
        mods.extend([
            x[:-3].split(".")[-1] for x in os.listdir(path)
            if x.endswith(".py") and not x.startswith("__")
           ])
    return sorted(mods)


def sums(checksum):
    logging.info(checksum)
    tbl = getmod("tbl")
    if not tbl:
        logging.info("no table")
        return
    if checksum and md5sum(tbl.__file__) != checksum:
        logging.info("table checksum error")
        return
    if "MD5" in dir(tbl):
        Mods.md5s.update(tbl.MD5)


def __dir__():
    return (
        'Mods',
        'getmod',
        'importer',
        'inits',
        'modules',
        'sums'
    )



# This file is placed in the Public Domain.


"web"


import logging
import os
import sys
import time


from http.server import HTTPServer, BaseHTTPRequestHandler


from tob.objects import Object
from tob.threads import launch


d = os.path.dirname
j = os.path.join


DEBUG = False
PATH = d(d(__file__))
PATH = j(PATH, "network", "html")


def init():
    if not os.path.exists(j(PATH, 'index.html')):
        logging.warning("no index.html")
        return
    try:
        server = HTTP((Cfg.hostname, int(Cfg.port)), HTTPHandler)
        server.start()
        logging.warning("http://%s:%s", Cfg.hostname, Cfg.port)
        return server
    except OSError as ex:
        logging.warning(str(ex))


class Cfg:

    hostname = "localhost"
    port     = 8000


class HTTP(HTTPServer, Object):

    daemon_thread = True
    allow_reuse_address = True

    def __init__(self, *args, **kwargs):
        HTTPServer.__init__(self, *args, **kwargs)
        Object.__init__(self)
        self.host = args[0]
        self._starttime = time.time()
        self._last = time.time()
        self._status = "start"

    def exit(self):
        time.sleep(0.2)
        self._status = ""
        self.shutdown()

    def start(self):
        launch(self.serve_forever)
        self._status = "ok"

    def request(self):
        self._last = time.time()

    def error(self, _request, _addr):
        exctype, excvalue, _trb = sys.exc_info()
        exc = exctype(excvalue)
        logging.exception(exc)


class HTTPHandler(BaseHTTPRequestHandler):

    def setup(self):
        BaseHTTPRequestHandler.setup(self)
        self._size = 0
        self._ip = self.client_address[0]

    def raw(self, data):
        self.wfile.write(data)

    def send(self, txt):
        self.wfile.write(bytes(txt, encoding="utf-8"))
        self.wfile.flush()

    def write_header(self, htype='text/plain', size=None):
        self.send_response(200)
        #self.send_header('Content-type', '%s; charset=%s ' % (htype, "utf-8"))
        self.send_header('Content-type', '%s;')
        if size is not None:
            self.send_header('Content-length', size)
        self.send_header('Server', "1")
        self.end_headers()

    def log(self, code):
        pass

    def do_GET(self):
        if "favicon" in self.path:
            return
        if DEBUG:
            return
        if self.path == "/":
            self.path = "index.html"
        self.path = PATH + os.sep + self.path
        if not os.path.exists(self.path):
            self.write_header("text/html")
            self.send_response(404)
            self.end_headers()
            return
        if "_images" in self.path:
            try:
                with open(self.path, "rb") as file:
                    img = file.read()
                    file.close()
                ext = self.path[-3]
                self.write_header(f"image/{ext}", len(img))
                self.raw(img)
            except (TypeError, FileNotFoundError, IsADirectoryError):
                self.send_response(404)
                self.end_headers()
            return
        try:
            with open(self.path, "r", encoding="utf-8", errors="ignore") as file:
                txt = file.read()
                file.close()
            self.write_header("text/html")
            self.send(txt)
        except (TypeError, FileNotFoundError, IsADirectoryError):
            self.send_response(404)
            self.end_headers()


def html2(txt):
    return """<!doctype html>
<html>
   %s
</html>
""" % txt



# This file is placed in the Public Domain.


"utilities"


import hashlib
import time


from .objects import Default, items


FORMATS = [
    "%Y-%M-%D %H:%M:%S",
    "%Y-%m-%d %H:%M:%S",
    "%Y-%m-%d",
    "%d-%m-%Y",
    "%d-%m",
    "%m-%d",
]


def elapsed(seconds, short=True):
    txt = ""
    nsec = float(seconds)
    if nsec < 1:
        return f"{nsec:.2f}s"
    yea     = 365 * 24 * 60 * 60
    week    = 7 * 24 * 60 * 60
    nday    = 24 * 60 * 60
    hour    = 60 * 60
    minute  = 60
    yeas    = int(nsec / yea)
    nsec   -= yeas * yea
    weeks   = int(nsec / week)
    nsec   -= weeks * week
    nrdays  = int(nsec / nday)
    nsec   -= nrdays * nday
    hours   = int(nsec / hour)
    nsec   -= hours * hour
    minutes = int(nsec / minute)
    nsec   -= int(minute * minutes)
    sec     = int(nsec)
    if yeas:
        txt += f"{yeas}y"
    if weeks:
        nrdays += weeks * 7
    if nrdays:
        txt += f"{nrdays}d"
    if short and txt:
        return txt.strip()
    if hours:
        txt += f"{hours}h"
    if minutes:
        txt += f"{minutes}m"
    if sec:
        txt += f"{sec}s"
    txt = txt.strip()
    return txt


def extract_date(daystr):
    daystr = daystr.encode('utf-8', 'replace').decode("utf-8")
    res = time.time()
    for fmat in FORMATS:
        try:
            res = time.mktime(time.strptime(daystr, fmat))
            break
        except ValueError:
            pass
    return res


def md5sum(path):
    with open(path, "r", encoding="utf-8") as file:
        txt = file.read().encode("utf-8")
        return hashlib.md5(txt).hexdigest()


def spl(txt):
    try:
        result = txt.split(",")
    except (TypeError, ValueError):
        result = [
            txt,
        ]
    return [x for x in result if x]


"methods"


def edit(obj, setter, skip=True):
    for key, val in items(setter):
        if skip and val == "":
            continue
        try:
            setattr(obj, key, int(val))
            continue
        except ValueError:
            pass
        try:
            setattr(obj, key, float(val))
            continue
        except ValueError:
            pass
        if val in ["True", "true"]:
            setattr(obj, key, True)
        elif val in ["False", "false"]:
            setattr(obj, key, False)
        else:
            setattr(obj, key, val)


def fmt(obj, args=[], skip=[], plain=False, empty=False):
    if not args:
        args = obj.__dict__.keys()
    txt = ""
    for key in args:
        if key.startswith("__"):
            continue
        if key in skip:
            continue
        value = getattr(obj, key, None)
        if value is None:
            continue
        if not empty and not value:
            continue
        if plain:
            txt += f"{value} "
        elif isinstance(value, str):
            txt += f'{key}="{value}" '
        elif isinstance(value, (int, float, dict, bool, list)):
            txt += f"{key}={value} "
        else:
            txt += f"{key}={name(value, True)} "
    return txt.strip()


def name(obj, short=False):
    typ = type(obj)
    res = ""
    if "__builtins__" in dir(typ):
        res = obj.__name__
    elif "__self__" in dir(obj):
        res = f"{obj.__self__.__class__.__name__}.{obj.__name__}"
    elif "__class__" in dir(obj) and "__name__" in dir(obj):
        res = f"{obj.__class__.__name__}.{obj.__name__}"
    elif "__class__" in dir(obj):
        res =  f"{obj.__class__.__module__}.{obj.__class__.__name__}"
    elif "__name__" in dir(obj):
        res = f"{obj.__class__.__name__}.{obj.__name__}"
    if short:
        res = res.split(".")[-1]
    return res


def parse(obj, txt):
    data = {
        "args": [],
        "cmd": "",
        "gets": Default(),
        "index": None,
        "init": "",
        "opts": "",
        "otxt": txt,
        "rest": "",
        "silent": Default(),
        "sets": Default(),
        "txt": ""
    }
    for k, v in data.items():
        setattr(obj, k, getattr(obj, k, v))
    args = []
    nr = -1
    for spli in txt.split():
        if spli.startswith("-"):
            try:
                obj.index = int(spli[1:])
            except ValueError:
                obj.opts += spli[1:]
            continue
        if "-=" in spli:
            key, value = spli.split("-=", maxsplit=1)
            obj.silent[key] = value
            obj.gets[key] = value
            continue
        if "==" in spli:
            key, value = spli.split("==", maxsplit=1)
            obj.gets[key] = value
            continue
        if "=" in spli:
            key, value = spli.split("=", maxsplit=1)
            obj.sets[key] = value
            continue
        nr += 1
        if nr == 0:
            obj.cmd = spli
            continue
        args.append(spli)
    if args:
        obj.args = args
        obj.txt  = obj.cmd or ""
        obj.rest = " ".join(obj.args)
        obj.txt  = obj.cmd + " " + obj.rest
    else:
        obj.txt = obj.cmd or ""


def __dir__():
    return (
        'elapsed',
        'extract_date',
        'name',
        'parse',
        'spl'
   )



# This file is placed in the Public Domain.


"persistence"


import datetime
import json
import os
import pathlib
import threading
import time


from .marshal import dump, load
from .objects import Object, deleted, fqn, search, update


lock = threading.RLock()


class Workdir:

    wdr = ""


class Cache:

    objs = {}

    @staticmethod
    def add(path, obj):
        Cache.objs[path] = obj

    @staticmethod
    def get(path):
        return Cache.objs.get(path, None)

    @staticmethod
    def update(path, obj):
        if path in Cache.objs:
            update(Cache.objs[path], obj)
        else:
            Cache.add(path, obj)


def cdir(path):
    pth = pathlib.Path(path)
    pth.parent.mkdir(parents=True, exist_ok=True)


def getpath(obj):
    return store(ident(obj))


def moddir():
    return os.path.join(Workdir.wdr, "mods")


def pidname(name):
    assert Workdir.wdr
    return os.path.join(Workdir.wdr, f"{name}.pid")


def skel():
    pth = pathlib.Path(store())
    pth.mkdir(parents=True, exist_ok=True)
    pth = pathlib.Path(moddir())
    pth.mkdir(parents=True, exist_ok=True)
    return str(pth)


def store(fnm=""):
    return os.path.join(Workdir.wdr, "store", fnm)


def types():
    return os.listdir(store())


"find"


def find(type=None, selector=None, removed=False, matching=False):
    if selector is None:
        selector = {}
    for pth in fns(type):
        obj = Cache.get(pth)
        if not obj:
            obj = Object()
            read(obj, pth)
            Cache.add(pth, obj)
        if not removed and deleted(obj):
            continue
        if selector and not search(obj, selector, matching):
            continue
        yield pth, obj


def fns(type=None):
    if type is not None:
        type = type.lower()
    path = store()
    for rootdir, dirs, _files in os.walk(path, topdown=True):
        for dname in dirs:
            if dname.count("-") != 2:
                continue
            ddd = os.path.join(rootdir, dname)
            if type and type not in ddd.lower():
                continue
            for fll in os.listdir(ddd):
                yield os.path.join(ddd, fll)


def fntime(daystr):
    datestr = " ".join(daystr.split(os.sep)[-2:])
    datestr = datestr.replace("_", " ")
    if "." in datestr:
        datestr, rest = datestr.rsplit(".", 1)
    else:
        rest = ""
    timed = time.mktime(time.strptime(datestr, "%Y-%m-%d %H:%M:%S"))
    if rest:
        timed += float("." + rest)
    return float(timed)



def last(obj, selector=None):
    if selector is None:
        selector = {}
    result = sorted(
                    find(fqn(obj), selector),
                    key=lambda x: fntime(x[0])
                   )
    res = ""
    if result:
        inp = result[-1]
        update(obj, inp[-1])
        res = inp[0]
    return res


"disk"


def ident(obj):
    return os.path.join(fqn(obj), *str(datetime.datetime.now()).split())


def read(obj, path):
    with lock:
        with open(path, "r", encoding="utf-8") as fpt:
            try:
                update(obj, load(fpt))
            except json.decoder.JSONDecodeError as ex:
                ex.add_note(path)
                raise ex


def write(obj, path=None):
    with lock:
        if path is None:
            path = getpath(obj)
        cdir(path)
        with open(path, "w", encoding="utf-8") as fpt:
            dump(obj, fpt, indent=4)
        Cache.update(path, obj)
        return path


def __dir__():
    return (
        'Cache',
        'Workdir',
        'cdir',
        'find',
        'fntime',
        'read',
        'skel',
        'types',
        'write'
    )



# This file is placed in the Public Domain.


"timers"


import datetime
import logging
import re
import time


from tob.brokers import Fleet
from tob.persist import find, write
from tob.utility import elapsed, extract_date
from tob.threads import Timed, launch


def init():
    nrs = 0
    for fnm, obj in find("timed"):
        if "time" not in dir(obj):
            continue
        nrs += 1
        diff = float(obj.time) - time.time()
        if diff > 0:
            timer = Timed(diff, Fleet.announce, obj.txt)
            timer.start()
        else:
            obj.__deleted__ = True
            write(obj, fnm)
    logging.warning("%s timers", nrs)


class NoDate(Exception):

    pass


def get_day(daystr):
    day = None
    month = None
    yea = None
    try:
        ymdre = re.search(r'(\d+)-(\d+)-(\d+)', daystr)
        if ymdre:
            (day, month, yea) = ymdre.groups()
    except ValueError:
        try:
            ymre = re.search(r'(\d+)-(\d+)', daystr)
            if ymre:
                (day, month) = ymre.groups()
                yea = time.strftime("%Y", time.localtime())
        except Exception as ex:
            raise NoDate(daystr) from ex
    if day:
        day = int(day)
        month = int(month)
        yea = int(yea)
        date = f"{day} {MONTHS[month]} {yea}"
        return time.mktime(time.strptime(date, r"%d %b %Y"))
    raise NoDate(daystr)


def get_hour(daystr):
    try:
        hmsre = re.search(r'(\d+):(\d+):(\d+)', str(daystr))
        hours = 60 * 60 * (int(hmsre.group(1)))
        hoursmin = hours  + int(hmsre.group(2)) * 60
        hmsres = hoursmin + int(hmsre.group(3))
    except AttributeError:
        pass
    except ValueError:
        pass
    try:
        hmre = re.search(r'(\d+):(\d+)', str(daystr))
        hours = 60 * 60 * (int(hmre.group(1)))
        hmsres = hours + int(hmre.group(2)) * 60
    except AttributeError:
        return 0
    except ValueError:
        return 0
    return hmsres


def get_time(txt):
    try:
        target = get_day(txt)
    except NoDate:
        target = to_day(today())
    hour =  get_hour(txt)
    if hour:
        target += hour
    return target


def parse_time(txt):
    seconds = 0
    target = 0
    txt = str(txt)
    for word in txt.split():
        if word.startswith("+"):
            seconds = int(word[1:])
            return time.time() + seconds
        if word.startswith("-"):
            seconds = int(word[1:])
            return time.time() - seconds
    if not target:
        try:
            target = get_day(txt)
        except NoDate:
            target = to_day(today())
        hour =  get_hour(txt)
        if hour:
            target += hour
    return target


def to_day(daystr):
    previous = ""
    line = ""
    daystr = str(daystr)
    res = None
    for word in daystr.split():
        line = previous + " " + word
        previous = word
        try:
            res = extract_date(line.strip())
            break
        except ValueError:
            res = None
        line = ""
    return res


def today():
    return str(datetime.datetime.today()).split()[0]


def tmr(event):
    result = ""
    if not event.rest:
        nmr = 0
        for _fn, obj in find('timed'):
            if "time" not in dir(obj):
                continue
            lap = float(obj.time) - time.time()
            if lap > 0:
                event.reply(f'{nmr} {obj.txt} {elapsed(lap)}')
                nmr += 1
        if not nmr:
            event.reply("no timers.")
        return result
    seconds = 0
    line = ""
    for word in event.args:
        if word.startswith("+"):
            try:
                seconds = int(word[1:])
            except (ValueError, IndexError):
                event.reply(f"{seconds} is not an integer")
                return result
        else:
            line += word + " "
    if seconds:
        target = time.time() + seconds
    else:
        try:
            target = get_day(event.rest)
        except NoDate:
            target = to_day(today())
        hour =  get_hour(event.rest)
        if hour:
            target += hour
    if not target or time.time() > target:
        event.reply("already passed given time.")
        return result
    diff = target - time.time()
    txt = " ".join(event.args[1:])
    timer = Timed(diff, Fleet.say, event.orig, event.channel, txt)
    timer.channel = event.channel
    timer.orig = event.orig
    timer.time = target
    timer.txt = txt
    write(timer)
    launch(timer.start)
    event.reply("ok " +  elapsed(diff))


"data"


MONTHS = [
    'Bo',
    'Jan',
    'Feb',
    'Mar',
    'Apr',
    'May',
    'Jun',
    'Jul',
    'Aug',
    'Sep',
    'Oct',
    'Nov',
    'Dec'
]



# This file is placed in the Public Domain.


"log text"


import time


from tob.objects import Object
from tob.persist import find, fntime, write
from tob.utility import elapsed


class Log(Object):

    def __init__(self):
        super().__init__()
        self.txt = ''


def log(event):
    if not event.rest:
        nmr = 0
        for fnm, obj in find('log'):
            lap = elapsed(time.time() - fntime(fnm))
            event.reply(f'{nmr} {obj.txt} {lap}')
            nmr += 1
        if not nmr:
            event.reply('no log')
        return
    obj = Log()
    obj.txt = event.rest
    write(obj)
    event.reply("ok")



# This file is placed in the Public Domain.
# ruff: noqa: F403,F405


"interface"


import logging
import sys
import unittest


import tob.brokers
import tob.clients
import tob.command
import tob.handler
import tob.logging
import tob.marshal
import tob.objects
import tob.package
import tob.persist
import tob.runtime
import tob.threads
import tob.utility


from tob.objects import *


import tob


TODO = tob


PACKAGE = [
    'brokers',
    'clients',
    'command',
    'handler',
    'logging',
    'marshal',
    'objects',
    'package',
    'persist',
    'runtime',
    'threads',
    'utility'
]


METHODS = [
    "__class__",
    "__delattr__",
    "__dict__",
    "__dir__",
    "__doc__",
    "__eq__",
    "__format__",
    "__ge__",
    "__getattribute__",
    "__getstate__",
    "__gt__",
    "__hash__",
    "__init__",
    "__init_subclass__",
    "__le__",
    "__len__",
    "__lt__",
    "__module__",
    "__ne__",
    "__new__",
    "__reduce__",
    "__reduce_ex__",
    "__repr__",
    "__setattr__",
    "__sizeof__",
    "__str__",
    "__subclasshook__",
    "__weakref__",
]


class TestInterface(unittest.TestCase):

    def test_package(self):
        okd = True
        for mod in PACKAGE:
            mod1 = getattr(TODO, mod, None)
            if not mod1:
                okd = False
                print(mod)
                break
        self.assertTrue(okd)

    def test_tob(self):
        okd = True
        obj = Object()
        dirr = dir(obj)
        for meth in METHODS:
            if meth not in dirr:
                okd = False
                print(f"{meth} not found")
        self.assertTrue(okd)


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger("SomeTest.testSomething").setLevel(logging.DEBUG)
    unittest.main()



# This file is placed in the Public Domain.


import unittest


from tob.marshal import dumps, loads
from tob.objects import Object, update


class TestTypes(unittest.TestCase):

    def test_dict(self):
        obj = loads(dumps({"a": "b"}))
        self.assertEqual(obj, {"a": "b"})

    def test_integer(self):
        obj = loads(dumps(1))
        self.assertEqual(obj, 1)

    def test_float(self):
        obj = loads(dumps(1.0))
        self.assertEqual(obj, 1.0)

    def test_string(self):
        obj = loads(dumps("test"))
        self.assertEqual(obj, "test")

    def test_true(self):
        obj = loads(dumps(True))
        self.assertEqual(obj, True)

    def test_false(self):
        obj = loads(dumps(False))
        self.assertEqual(obj, False)

    def test_object(self):
        ooo = Object()
        ooo.a = "b"
        obj = Object()
        update(obj, loads(dumps(ooo)))
        self.assertEqual(obj.a, "b")



# This file is placed in the Public Domain.


import unittest


from tob.marshal import dumps, loads
from tob.objects import Object


class TestDecoder(unittest.TestCase):

    def test_loads(self):
        obj = Object()
        obj.test = "bla"
        oobj = loads(dumps(obj))
        self.assertEqual(oobj["test"], "bla")

    def test_doctest(self):
        self.assertTrue(__doc__ is None)



# This file is placed in the Public Domain.


"uptime"


import time


from tob.runtime import STARTTIME
from tob.utility import elapsed


def upt(event):
    event.reply(elapsed(time.time()-STARTTIME))



# This file is placed in the Public Domain.


"enable silence mode"


from tob.brokers import Fleet


def sil(event):
    bot = Fleet.get(event.orig)
    bot.silent = True
    event.reply("ok")


def lou(event):
    bot = Fleet.get(event.orig)
    bot.silent = False
    event.reply("ok")



# This file is placed in the Public Domain.


"run non-blocking"


import logging
import queue
import threading
import time
import _thread


from .utility import name


class Thread(threading.Thread):

    def __init__(self, func, *args, daemon=True, **kwargs):
        super().__init__(None, self.run, None, (), daemon=daemon)
        self.name = kwargs.get("name", name(func))
        self.queue = queue.Queue()
        self.result = None
        self.starttime = time.time()
        self.stopped = threading.Event()
        self.queue.put((func, args))

    def __iter__(self):
        return self

    def __next__(self):
        yield from dir(self)

    def join(self, timeout=None):
        result = None
        try:
            super().join(timeout)
            result = self.result
        except (KeyboardInterrupt, EOFError):
            _thread.interrupt_main()
        return result

    def run(self):
        func, args = self.queue.get()
        try:
            self.result = func(*args)
        except (KeyboardInterrupt, EOFError):
            _thread.interrupt_main()
        except Exception as ex:
            logging.exception(ex)
            _thread.interrupt_main()


class Timy(threading.Timer):

    def __init__(self, sleep, func, *args, **kwargs):
        super().__init__(sleep, func)
        self.name = kwargs.get("name", name(func))
        self.sleep = sleep
        self.state = {}
        self.state["latest"] = time.time()
        self.state["starttime"] = time.time()
        self.starttime = time.time()


class Timed:

    def __init__(self, sleep, func, *args, thrname="", **kwargs):
        self.args = args
        self.func = func
        self.kwargs = kwargs
        self.sleep = sleep
        self.name = thrname or kwargs.get("name", name(func))
        self.target = time.time() + self.sleep
        self.timer = None

    def run(self):
        self.timer.latest = time.time()
        self.func(*self.args)

    def start(self):
        self.kwargs["name"] = self.name
        timer = Timy(self.sleep, self.run, *self.args, **self.kwargs)
        timer.start()
        self.timer = timer

    def stop(self):
        if self.timer:
            self.timer.cancel()


class Repeater(Timed):

    def run(self):
        launch(self.start)
        super().run()


def launch(func, *args, **kwargs):
    thread = Thread(func, *args, **kwargs)
    thread.start()
    return thread


def __dir__():
    return (
        'Repeater',
        'Thread',
        'Timed',
        'launch'
   )



# This file is placed in the Public Domain.


"logging at level"


import logging as log


LEVELS = {
    'debug': log.DEBUG,
    'info': log.INFO,
    'warning': log.WARNING,
    'warn': log.WARNING,
    'error': log.ERROR,
    'critical': log.CRITICAL
}


datefmt = "%H:%M:%S"
format_short = "%(module)-3s %(message)-76s"


class Formatter(log.Formatter):

    def format(self, record):
        record.module = record.module.upper()
        return log.Formatter.format(self, record)


def level(loglevel="debug"):
    if loglevel != "none":
        ch = log.StreamHandler()
        formatter = Formatter(fmt=format_short, datefmt=datefmt)
        ch.setFormatter(formatter)
        lvl = LEVELS.get(loglevel)
        if not lvl:
            return
        logger = log.getLogger()
        logger.setLevel(LEVELS.get(loglevel))
        logger.addHandler(ch)


def __dir__():
    return (
        'level',
   )



# This file is placed in the Public Domain.


"UDP to IRC relay"


import logging
import select
import socket
import sys
import threading
import time


from tob.brokers import Fleet
from tob.objects import Object
from tob.threads import launch


DEBUG = False


def init():
    udpd = UDP()
    udpd.start()
    logging.warning("http://%s:%s", Cfg.host, Cfg.port)
    return udpd


class Cfg(Object):

    addr = ""
    host = "localhost"
    port = 5500


class UDP(Object):

    def __init__(self):
        Object.__init__(self)
        self.stopped = False
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self._sock.setblocking(1)
        self._starttime = time.time()
        self.ready = threading.Event()

    def output(self, txt, addr=None):
        if addr:
            Cfg.addr = addr
        Fleet.announce(txt.replace("\00", ""))

    def loop(self):
        try:
            self._sock.bind((Cfg.host, Cfg.port))
        except socket.gaierror:
            return
        self.ready.set()
        while not self.stopped:
            (txt, addr) = self._sock.recvfrom(64000)
            if self.stopped:
                break
            data = str(txt.rstrip(), "utf-8")
            if not data:
                break
            self.output(data, addr)

    def exit(self):
        self.stopped = True
        self._sock.settimeout(0.01)
        self._sock.sendto(
                          bytes("exit", "utf-8"),
                          (Cfg.host, Cfg.port)
                         )

    def start(self):
        launch(self.loop)


def toudp(host, port, txt):
    if DEBUG:
        return
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.sendto(bytes(txt.strip(), "utf-8"), (host, port))


def udp(event):
    if event.rest:
        toudp(Cfg.host, Cfg.port, event.rest)
        return
    if not select.select(
                         [sys.stdin, ],
                         [],
                         [],
                         0.0
                        )[0]:
        event.reply("udp <text>")
        return
    size = 0
    while 1:
        try:
            (inp, _out, err) = select.select(
                                             [sys.stdin,],
                                             [],
                                             [sys.stderr,]
                                            )
        except KeyboardInterrupt:
            return
        if err:
            break
        stop = False
        for sock in inp:
            txt = sock.readline()
            if not txt:
                stop = True
                break
            size += len(txt)
            toudp(Cfg.host, Cfg.port, txt)
        if stop:
            break



# This file is placed in the Public Domain.


"REST"


import logging
import os
import sys
import time


from http.server import HTTPServer, BaseHTTPRequestHandler


from tob.objects import Object
from tob.persist import store, types
from tob.threads import launch


DEBUG = False


def init():
    try:
        rest = REST((Config.hostname, int(Config.port)), RESTHandler)
        rest.start()
        logging.warning("http://%s:%s", Config.hostname, Config.port)
        return rest
    except OSError as ex:
        logging.error(str(ex))


class Config:

    hostname = "localhost"
    port     = 10102


class REST(HTTPServer, Object):

    allow_reuse_address = True
    daemon_thread = True

    def __init__(self, *args, **kwargs):
        HTTPServer.__init__(self, *args, **kwargs)
        Object.__init__(self)
        self.host = args[0]
        self._last = time.time()
        self._starttime = time.time()
        self._status = "start"

    def exit(self):
        self._status = ""
        time.sleep(0.2)
        self.shutdown()

    def start(self):
        self._status = "ok"
        launch(self.serve_forever)

    def request(self):
        self._last = time.time()

    def error(self, _request, _addr):
        exctype, excvalue, _trb = sys.exc_info()
        exc = exctype(excvalue)
        logging.exception(exc)


class RESTHandler(BaseHTTPRequestHandler):

    def setup(self):
        BaseHTTPRequestHandler.setup(self)
        self._ip = self.client_address[0]
        self._size = 0

    def send(self, txt):
        self.wfile.write(bytes(txt, "utf-8"))
        self.wfile.flush()

    def write_header(self, htype='text/plain'):
        self.send_response(200)
        self.send_header('Content-type', '%s; charset=%s ' % (htype, "utf-8"))
        self.send_header('Server', "1")
        self.end_headers()

    def do_GET(self):
        if DEBUG:
            return
        if "favicon" in self.path:
            return
        if self.path == "/":
            self.write_header("text/html")
            txt = ""
            for fnm in types():
                txt += f'<a href="http://{Config.hostname}:{Config.port}/{fnm}">{fnm}</a><br>\n'
            self.send(html(txt.strip()))
            return
        fnm = store() + self.path
        fnm = os.path.abspath(fnm)
        if os.path.isdir(fnm):
            self.write_header("text/html")
            txt = ""
            for fnn in os.listdir(fnm):
                filename = self.path  + os.sep + fnn
                txt += f'<a href="http://{Config.hostname}:{Config.port}/{filename}">{filename}</a><br>\n'
            self.send(txt.strip())
            return
        try:
            with open(fnm, "r", encoding="utf-8") as file:
                txt = file.read()
                file.close()
            self.write_header("text/html")
            self.send(html(txt))
        except (TypeError, FileNotFoundError, IsADirectoryError) as ex:
            self.send_response(404)
            logging.exception(ex)
            self.end_headers()

    def log(self, code):
        pass


def html(txt):
    return """<!doctype html>
<html>
   %s
</html>
""" % txt



# This file is placed in the Public Domain.


"a clean namespace"


class Object:

    def __contains__(self, key):
        return key in dir(self)

    def __getitem__(self, key):
        return self.__dict__.get(key)

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __str__(self):
        return str(self.__dict__)


class Default(Object):

    def __getattr__(self, key):
        return self.__dict__.get(key, "")


def construct(obj, *args, **kwargs):
    if args:
        val = args[0]
        if isinstance(val, zip):
            update(obj, dict(val))
        elif isinstance(val, dict):
            update(obj, val)
        else:
            update(obj, vars(val))
    if kwargs:
        update(obj, kwargs)


def deleted(obj):
    return "__deleted__" in dir(obj) and obj.__deleted__


def fqn(obj):
    kin = str(type(obj)).split()[-1][1:-2]
    if kin == "type":
        kin = f"{obj.__module__}.{obj.__name__}"
    return kin


def items(obj):
    if isinstance(obj, dict):
        return obj.items()
    return obj.__dict__.items()


def keys(obj):
    if isinstance(obj, dict):
        return obj.keys()
    return obj.__dict__.keys()


def search(obj, selector, matching=False):
    res = False
    for key, value in items(selector):
        val = getattr(obj, key, None)
        if not val:
            continue
        if matching and value == val:
            res = True
        elif str(value).lower() in str(val).lower():
            res = True
        else:
            res = False
            break
    return res


def update(obj, data, empty=True):
    for key, value in items(data):
        if not empty and not value:
            continue
        setattr(obj, key, value)


def values(obj):
    if isinstance(obj, dict):
        return obj.values()
    return obj.__dict__.values()


def __dir__():
    return (
        'Default',
        'Object',
        'construct',
        'deleted',
        'fqn',
        'items',
        'keys',
        'update',
        'values'
    )



# This file is placed in the Public Domain.


import unittest


from tob.marshal import dumps
from tob.objects import Object


VALIDJSON = '{"test": "bla"}'


class TestEncoder(unittest.TestCase):

    def test_dumps(self):
        obj = Object()
        obj.test = "bla"
        self.assertEqual(dumps(obj), VALIDJSON)



# This file is placed in the Public Domain.


"manual"


TXT = """
N A M E


    TOB - bot in reverse !


S Y N O P S I S


    tob <cmd> [key=val] [key==val]
    tob -cvaw [init=mod1,mod2]
    tob -d
    tob -s


D E S C R I P T I O N


    TOB has all you need to program a unix cli program, such as disk
    perisistence for configuration files, event handler to handle the
    client/server connection, easy programming of your own commands, etc.

    TOB contains python3 code to program objects in a functional way.
    it provides an "clean namespace" Object class that only has dunder
    methods, so the namespace is not cluttered with method names. This
    makes storing and reading to/from json possible.

    TOB is a python3 IRC bot, it can connect to IRC, fetch and
    display RSS feeds, take todo notes, keep a shopping list and log
    text. You can run it under systemd for 24/7 presence in a IRC channel.

    TOB is Public Domain.


I N S T A L L


    installation is done with pipx

    $ pipx install tob
    $ pipx ensurepath

    <new terminal>

    $ tob srv > tob.service
    $ sudo mv tob.service /etc/systemd/system/
    $ sudo systemctl enable tob --now

    joins #tob on localhost


U S A G E 


    use tob to control the program, default it does nothing

    $ tob
    $

    see list of commands


    $ tob cmd
    cfg,cmd,dne,dpl,err,exp,imp,log,mod,mre,nme,
    pwd,rem,req,res,rss,srv,syn,tdo,thr,upt


    start console

    $ tob -c


    start console and run irc and rss 

    $ tob -c init=irc,rss

    list available modules

    $ tob mod
    err,flt,fnd,irc,llm,log,mbx,mdl,mod,req,rss,
    rst,slg,tdo,thr,tmr,udp,upt``

    start daemon

    $ tob -d
    $

    start service

    $ tob -s

    <runs until ctrl-c>


C O M M A N D S


    here is a list of available commands

    cfg - irc configuration
    cmd - commands
    dpl - sets display items
    err - show errors
    exp - export opml (stdout)
    imp - import opml
    log - log text
    mre - display cached output
    pwd - sasl nickserv name/pass
    rem - removes a rss feed
    res - restore deleted feeds
    rss - add a feed
    syn - sync rss feeds
    tdo - add todo item
    thr - show running threads
    upt - show uptime


C O N F I G U R A T I O N

    irc

    $ tob cfg server=<server>
    $ tob cfg channel=<channel>
    $ tob cfg nick=<nick>

    sasl

    $ tob pwd <nsnick> <nspass>
    $ tob cfg password=<frompwd>

    rss

    $ tob rss <url>
    $ tob dpl <url> <item1,item2>
    $ tob rem <url>
    $ tob nme <url> <name>

    opml

    $ tob exp
    $ tob imp <filename>


P R O G R A M M I N G


    tob has it's modules in the ~/.tob/mods directory so for a hello world
    command you would  edit a file in ~/.tob/mods/hello.py and add the
    following


    def hello(event):
        event.reply("hello world !!")


    typing the hello command would result into a nice hello world !!


    $ tob hello
    hello world !!


    commands run in their own thread and the program borks on exit to enable a
    short debug cycle, output gets flushed on print so exceptions appear in the
    systemd logs. modules can contain your own written python3 code.


F I L E S


    ~/.tob
    ~/.local/bin/tob
    ~/.local/pipx/venvs/tob/*


A U T H O R


    Bart Thate <bthate@dds.nl>


C O P Y R I G H T


    TOB is Public Domain
"""


def man(event):
    event.reply(TXT)



# This file is placed in the Public Domain.


"client for a string"


class Fleet:

    clients = {}

    @staticmethod
    def add(client):
        Fleet.clients[repr(client)] = client

    @staticmethod
    def all():
        return Fleet.clients.values()

    @staticmethod
    def announce(txt):
        for client in Fleet.all():
            client.announce(txt)

    @staticmethod
    def display(evt):
        client = Fleet.get(evt.orig)
        client.display(evt)

    @staticmethod
    def get(orig):
        return Fleet.clients.get(orig, None)

    @staticmethod
    def say(orig, channel, txt):
        client = Fleet.get(orig)
        client.say(channel, txt)

    @staticmethod
    def shutdown():
        for client in Fleet.all():
            client.wait()
            client.stop()


def __dir__():
    return (
        'Fleet',
   )



# This file is placed in the Public Domain.


"runtime"


import inspect
import os
import pathlib
import sys
import time


from .command import parse, table
from .logging import level
from .package import Mods, modules, sums
from .persist import Workdir, moddir, skel


STARTTIME = time.time()


class Config:

    debug = False
    default = "irc,rss"
    init  = ""
    level = "warn"
    name = os.path.dirname(__file__).split(os.sep)[-1]
    opts = ""
    verbose = False
    version = 136


def boot(mods, checksum, doparse=True):
    Mods.add("modules", os.path.dirname(inspect.getfile(mods)))
    Mods.add("mods", moddir())
    if doparse:
        parse(Config, " ".join(sys.argv[1:]))
        Config.level = Config.sets.level or Config.level
    Workdir.wdr = Workdir.wdr or os.path.expanduser(f"~/.{Config.name}")
    level(Config.level)
    if "a" in Config.opts:
        Config.sets.init = ",".join(modules())
    skel()
    table()
    sums(checksum)


def daemon(verbose=False):
    pid = os.fork()
    if pid != 0:
        os._exit(0)
    os.setsid()
    pid2 = os.fork()
    if pid2 != 0:
        os._exit(0)
    if not verbose:
        with open('/dev/null', 'r', encoding="utf-8") as sis:
            os.dup2(sis.fileno(), sys.stdin.fileno())
        with open('/dev/null', 'a+', encoding="utf-8") as sos:
            os.dup2(sos.fileno(), sys.stdout.fileno())
        with open('/dev/null', 'a+', encoding="utf-8") as ses:
            os.dup2(ses.fileno(), sys.stderr.fileno())
    os.umask(0)
    os.chdir("/")
    os.nice(10)


def forever():
    while True:
        try:
            time.sleep(0.1)
        except (KeyboardInterrupt, EOFError):
            break


def pidfile(filename):
    if os.path.exists(filename):
        os.unlink(filename)
    path2 = pathlib.Path(filename)
    path2.parent.mkdir(parents=True, exist_ok=True)
    with open(filename, "w", encoding="utf-8") as fds:
        fds.write(str(os.getpid()))


def privileges():
    import getpass
    import pwd
    pwnam2 = pwd.getpwnam(getpass.getuser())
    os.setgid(pwnam2.pw_gid)
    os.setuid(pwnam2.pw_uid)


def wrapped(func):
    try:
        func()
    except (KeyboardInterrupt, EOFError):
        pass


def wrap(func):
    import termios
    old = None
    try:
        old = termios.tcgetattr(sys.stdin.fileno())
    except termios.error:
        pass
    try:
        wrapped(func)
    finally:
        if old:
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, old)


def __dir__():
    return (
        'STARTTIME',
        'Config',
        'boot',
        'daemon',
        'forever',
        'pidfile',
        'privileges',
        'wrap',
        'wrapped'
    )



# This file is placed in the Public Domain.


"find"


import time


from tob.persist import find, fntime, types
from tob.utility import elapsed, fmt


def fnd(event):
    if not event.rest:
        res = sorted([x.split('.')[-1].lower() for x in types()])
        if res:
            event.reply(",".join(res))
        else:
            event.reply("no data yet.")
        return
    otype = event.args[0]
    nmr = 0
    for fnm, obj in list(find(otype, event.gets)):
        event.reply(f"{nmr} {fmt(obj)} {elapsed(time.time()-fntime(fnm))}")
        nmr += 1
    if not nmr:
        event.reply("no result")



# This file is placed in the Public Domain.


import unittest


from tob.marshal import dumps, loads
from tob.objects import Object, update


class TestComposite(unittest.TestCase):

    def testcomposite(self):
        obj = Object()
        obj.obj = Object()
        obj.obj.a = "test"
        self.assertEqual(obj.obj.a, "test")

    def testcompositeprint(self):
        obj = Object()
        obj.obj = Object()
        obj.obj.a = "test"
        txt = dumps(obj)
        ooo = Object()
        update(ooo, loads(txt))
        self.assertTrue(ooo.obj)



# This file is placed in the Public Domain.


import unittest


from tob.objects import Object


class A(Object):

    pass

class B(Object):

    pass


class C(A, B):

    bla = "mekker"


class D:

    pass


class E(A, D):

    pass


class F(C, D):

    pass


class H(dict):

    pass


#class I(object):

#    pass


#class J(A, I):

#    pass


#class K(J, H):

#    pass


class TestInherit(unittest.TestCase):

    def testinherit1(self):
        c = C()
        self.assertEqual(type(c), C)

    def testinherit2(self):
        e = E()
        self.assertEqual(type(e), E)

    def testinherit3(self):
        f = F()
        self.assertEqual(type(f), F)

    def testinherit4(self):
        f = F()
        self.assertEqual(f.bla, "mekker")

    def testinherit5(self):
        h = H()
        self.assertEqual(type(h), H)
